/*
	
	TITLE:		CGraphics.cpp
	DESC:		Interface implementation for shadows sample program
	DATE:		August 2003

	AUTHOR:		Jack Hoxley

*/

//include system includes
#include <math.h>
#include <Windows.h>
#include <exception>
#include <d3dx9.h>
#include <stdio.h>

//include the implementation header
#include "CGraphics.h"
#include "CGraphicsException.h"
#include "ShadowCalculator.h"

//conditional statements:
//#define DISPLAY_VSYNC							//comment out this #define to disable vsync'ing.
#define AMB_LIGHT D3DCOLOR_XRGB( 64, 64, 64 )	//use this to easily change the ambient lighting across the whole application
#define COMBINE_FIRST_2_PASSES					//determines whether we combine the z-fill and ambient passes
//#define RENDER_VOLUMES						//displays the shadow volumes on screen
#define SHADOW_VOLUME_UPDATE_ALTERNATE			//updates the shadow volumes using a per-frame delta..


/*-------------------------------------------------------------*\

	CGraphics(HWND)
	---------------

	hWnd =	valid handle to a win32 window that is the target
			of 3D rendering

	Constructor for a windowed mode instance of the 3D engine

\*-------------------------------------------------------------*/
CGraphics::CGraphics(HWND hWnd) {
	
	ResetAllVars();

	InitialiseDevice(hWnd,0,0,0,true);

} //ctor(hwnd)


/*-------------------------------------------------------------*\

	CGraphics(HWND,INT,INT,INT)
	---------------------------

	hWnd =	Valid handle to a win32 render target
	iW   =	Width of the screen
	iH   =	Height of the screen
	iD   =	Depth of the display (16,24 or 32 only)

	Creates a fullscreen instance of the graphics engine, 
	the engine will check compatability before allowing the
	initialisation to properly occur.

\*-------------------------------------------------------------*/
CGraphics::CGraphics(HWND hWnd, int iW, int iH, int iD) {
	
	ResetAllVars();
	
	if (IsDisplayModeOkay(iW,iH,iD)) {
		//hide the mouse cursor while in fullscreen mode
		do {} while (ShowCursor(false)>0);

		//set the ball rolling...
		InitialiseDevice(hWnd,iW,iH,iD,false);

	} else
		throw CGraphicsException("Resolution is not supported!","CGraphics.cpp",__LINE__);

} //ctor(hwnd,int,int,int)


/*-------------------------------------------------------------*\

	~CGraphics()
	------------

	Standard destructor for the class, makes sure that all
	memory is deallocated and that any necessary objects
	are terminated in the correct manner.

\*-------------------------------------------------------------*/
CGraphics::~CGraphics() {

	//free other memory
	SAFE_DELETE( sDispInfo );
	SAFE_DELETE( sDevInfo );
	SAFE_DELETE( lList[0] );
	SAFE_DELETE( lList[1] );
	SAFE_DELETE( lList[2] );
	SAFE_DELETE( lList[3] );

	//release the shadows
	for ( int i = 0; i < 12; i++ )
		SAFE_DELETE( pVols[i] );

	//release the geometry
	SAFE_DELETE( mshCaster[0] );
	SAFE_DELETE( mshCaster[1] );

	//release textures
	SAFE_RELEASE( pLightTex );

	//free the device
	SAFE_RELEASE( pDev );
	
	//free the D3D9 interface
	SAFE_RELEASE( pD3D9 );

	//show the mouse again
	do {} while (ShowCursor(true)<=0);
} //~CGraphics()


/*-------------------------------------------------------------*\

	ResetAllVars()
	--------------

	makes sure that all internal members have the correct
	default values. Intended to be used when the application
	starts, although could be used at any point really..

\*-------------------------------------------------------------*/
void CGraphics::ResetAllVars( ) {

	//configure default values
	bWireFrame = false;
	bUseTextures = true;
	bCulling = true;
	bDrawFPS = true;
	bAnimateCamera = true;
	iOutputLevel = 0;
	iModelDetail = 0;
	bDisplayStats = false;
	iShadowUpdateSpeed = 0;

	bInitOkay = false;
	fCurrFPS = 0.0;
	iCurrCnt = 0;
	dwLastCameraChange = GetTickCount();
	fCamAngle = 0.0f;

	for ( int i = 0; i < ( iLightCount * 4 ); i++ )
		pVols[ i ] = new ShadowVolume(); //allocate the memory

	lList[0] = NULL;
	lList[1] = NULL;
	lList[2] = NULL;
	lList[3] = NULL;

	pLightTex = NULL;
	pD3D9 = NULL;
	pDev = NULL;

	for ( int i = 0; i < 8; i++ )
		mshCaster[ i ] = NULL;

} //ResetAllVars()


/*-------------------------------------------------------------*\

	isDisplayModeOkay(int,int,int)
	------------------------------

	iW = Width of the screen
	iH = Height of the screen
	iD = Colour depth (16,24 or 32)

	A static function that allows an application to test the
	specified display mode BEFORE it attempts to create the
	graphics engine with the given paramters.

\*-------------------------------------------------------------*/
bool CGraphics::IsDisplayModeOkay(int iW, int iH, int iD) {
	LPDIRECT3D9 pD3D;
	D3DFORMAT fmtTmp;

	pD3D = Direct3DCreate9( D3D_SDK_VERSION );
	if ( NULL == pD3D )
		return false;

	switch (iD) {
		case 16:
			fmtTmp = D3DFMT_R5G6B5;
			break;
		case 24:
			fmtTmp = D3DFMT_R8G8B8;
			break;
		case 32:
			fmtTmp = D3DFMT_X8R8G8B8;
			break;
		default:
			throw CGraphicsException("Invalid display depth specified","CGraphics.cpp",__LINE__);
	}
	
	UINT cnt = pD3D->GetAdapterModeCount( D3DADAPTER_DEFAULT, fmtTmp);

	for (UINT i = 0; i < cnt; i++ ) {
		D3DDISPLAYMODE Mode;

		pD3D->EnumAdapterModes( D3DADAPTER_DEFAULT, fmtTmp, i, &Mode);

		if ( (Mode.Width == iW) && (Mode.Height == iH) ) {
			pD3D->Release();
			return true;
		}
	}
	
	pD3D->Release();
	return false;
}


/*-------------------------------------------------------------*\

	OneFrameRender()
	----------------

	Called once on each frame, its job is to take all active
	geometry and render it accordingly on the screen. In the
	case of the shadows, the rendering passes are contained here
	but calls to generate supporting shadow volumes are in
	OneFrameUpdate()..

\*-------------------------------------------------------------*/
bool CGraphics::OneFrameRender() {
	//dont do any work if the device was not configured.
	if (!bInitOkay)
		return false;

	//0. INITIALISE ANY VARIABLES
	int iTrisDrawn = 0; //simple statistic

	//1. RENDER THE ACTUAL SCENE TO THE TEXTURE

		//clear any existing colour and depth information.
		//MUST CLEAR to BLACK (0,0,0) to stop it interfering with additive lighting
		pDev->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0f, 0 );

		//tell D3D we're gonna start rendering
		pDev->BeginScene();

#ifdef COMBINE_FIRST_2_PASSES
		if ( bAmbientLight ) {
			//colour buffer ON
			pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
			pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ONE );
			pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
			//ambient lighting ON
			pDev->SetRenderState( D3DRS_LIGHTING, TRUE );
			pDev->SetRenderState( D3DRS_AMBIENT, AMB_LIGHT );
			pDev->LightEnable( 0, FALSE );
			pDev->LightEnable( 1, FALSE );
			pDev->LightEnable( 2, FALSE );
			pDev->LightEnable( 3, FALSE );

		} else {	
			//colour buffer OFF
			pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
			pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ZERO );
			pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ZERO );
			//lighting OFF
			pDev->SetRenderState( D3DRS_LIGHTING, FALSE );

		} //if(bAmbientLight)

		//depth buffer ON (write+test)
		pDev->SetRenderState( D3DRS_ZENABLE, TRUE );
		pDev->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );
		//stencil buffer OFF
		pDev->SetRenderState( D3DRS_STENCILENABLE, FALSE );

		//FIRST PASS: render scene to the depth buffer
		for ( int i = 0; i < 8; i++ ) {
			mshCaster[i]->Render( pDev );
			iTrisDrawn += mshCaster[i]->trianglesInMesh();
		}

		//OPTIMISED: all subsequent rendering is subject to this
		pDev->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

#else

		//A. Render the first pass:
			//colour buffer OFF
			pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
			pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ZERO );
			pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ZERO );
			//lighting OFF
			pDev->SetRenderState( D3DRS_LIGHTING, FALSE );
			//depth buffer ON (write+test)
			pDev->SetRenderState( D3DRS_ZENABLE, TRUE );
			pDev->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );
			//stencil buffer OFF
			pDev->SetRenderState( D3DRS_STENCILENABLE, FALSE );

			//FIRST PASS: render scene to the depth buffer
			for ( int i = 0; i < 8; i++ ) {
				mshCaster[i]->Render( pDev );
				iTrisDrawn += mshCaster[i]->trianglesInMesh();
			}

			//OPTIMISED: all subsequent rendering is subject to this
			pDev->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

		//B. Render the second pass:
			if ( bAmbientLight ) {
				//colour buffer ON
				pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
				pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ONE );
				pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
				//ambient lighting ON
				pDev->SetRenderState( D3DRS_LIGHTING, TRUE );
				pDev->SetRenderState( D3DRS_AMBIENT, AMB_LIGHT );
				pDev->LightEnable( 0, FALSE );
				pDev->LightEnable( 1, FALSE );
				pDev->LightEnable( 2, FALSE );
				pDev->LightEnable( 3, FALSE );

				//SECOND PASS: render scene to the colour buffer
				for ( int i = 0; i < 8; i++ ) {
					mshCaster[i]->Render( pDev );
					iTrisDrawn += mshCaster[i]->trianglesInMesh();
				}

			} //if(bAmbientLight)
#endif

			//PROFILE: fMS_RenderShadows
			QueryPerformanceCounter( &liStart );

		//C. Render all subsequent passes
			//colour buffer OFF
			//ambient lighting OFF
			//depth buffer ON (test ONLY)
			//stencil buffer OFF
			//for each light
			int iSVolIdx = 0;
			for ( int i = 0; i < iLightCount; i ++ ) {

				//if  enabled && casting shadows
				if ( lList[i]->bEnabled && lList[i]->bCastsShadows ) {

					//clear stencil buffer
					pDev->Clear( 0, NULL, D3DCLEAR_STENCIL, D3DCOLOR_XRGB(0,0,0), 1.0f, 0 );

					//turn OFF colour buffer
					pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
					pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ZERO );
					pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );

					//disable lighting (not needed for stencil writes!)
					pDev->SetRenderState( D3DRS_LIGHTING, FALSE );

					//turn ON stencil buffer
					pDev->SetRenderState( D3DRS_STENCILENABLE, TRUE );
					pDev->SetRenderState( D3DRS_STENCILFUNC,  D3DCMP_ALWAYS );

					//render shadow volume
						//OPTIMISED: use support for 2-sided stencil
						if ( b2SidedStencils ) {
							//USE THE LATEST 1-PASS METHOD
							
							//configure the necessary render states
							pDev->SetRenderState( D3DRS_STENCILPASS,		D3DSTENCILOP_INCR );
							pDev->SetRenderState( D3DRS_CULLMODE,			D3DCULL_NONE );
							pDev->SetRenderState( D3DRS_TWOSIDEDSTENCILMODE, TRUE );

							//render the geometry once to the stencil buffer
							for ( int j = iSVolIdx; j < ( iSVolIdx + 4 ); j++ ) {
								pVols[j]->Render( pDev );
								iTrisDrawn += pVols[j]->dwTriangles();
							}

							//reset any necessary states
							pDev->SetRenderState( D3DRS_TWOSIDEDSTENCILMODE, FALSE );


						} else {
							//USE THE TRADITIONAL 2-PASS METHOD

							//set stencil to increment
							pDev->SetRenderState( D3DRS_STENCILPASS,      D3DSTENCILOP_INCR );
								//render front faces
								pDev->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
								for ( int j = iSVolIdx; j < ( iSVolIdx + 4 ); j++ ) {
									pVols[j]->Render( pDev );
									iTrisDrawn += pVols[j]->dwTriangles();
								}
							//set stencil to decrement
							pDev->SetRenderState( D3DRS_STENCILPASS,      D3DSTENCILOP_DECR );
								//render back faces
								pDev->SetRenderState( D3DRS_CULLMODE, D3DCULL_CW );
								for ( int j = iSVolIdx; j < ( iSVolIdx + 4 ); j++ ) {
									pVols[j]->Render( pDev );
									iTrisDrawn += pVols[j]->dwTriangles();
								}
								pDev->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
						}

			#ifdef RENDER_VOLUMES
						for ( int j = iSVolIdx; j < ( iSVolIdx + 4 ); j++ ) {
							pVols[j]->Render( pDev );
							iTrisDrawn += pVols[j]->dwTriangles();
						}
			#endif

						//Increment shadow volume index, slightly hack-ish but it'll do.
						//basically, there is no formula for the idx into pVols[] so we
						//just have to keep on counting...
						iSVolIdx += 4;

					//alter stencil buffer
					pDev->SetRenderState( D3DRS_STENCILFUNC, D3DCMP_GREATER );
					pDev->SetRenderState( D3DRS_STENCILPASS, D3DSTENCILOP_KEEP );

					//turn on CURRENT light, turn off all others
					pDev->SetRenderState( D3DRS_LIGHTING, TRUE );
					pDev->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_XRGB(0,0,0) ); //turn off ambient light
					pDev->LightEnable( 0, i == 0 ? TRUE : FALSE );
					pDev->LightEnable( 1, i == 1 ? TRUE : FALSE );
					pDev->LightEnable( 2, i == 2 ? TRUE : FALSE );
					pDev->LightEnable( 3, i == 3 ? TRUE : FALSE );
					
					//turn ON colour buffer
					pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
					pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_ONE );
					pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
					
					//render scene
					for ( int k = 0; k < 8; k++ ) {
						mshCaster[k]->Render( pDev );
						iTrisDrawn += mshCaster[k]->trianglesInMesh();
					}

					//reset any necessary render states
					pDev->SetRenderState( D3DRS_STENCILENABLE, FALSE );

				} //if(enabled&&caster)

			} //for(each light)


			QueryPerformanceCounter( &liEnd );
			QueryPerformanceFrequency( &liFreq );
			double dFinal = (double)( liEnd.QuadPart - liStart.QuadPart ) / (double)liFreq.QuadPart;
			fMS_RenderShadows = (float)( dFinal * 1000 );

		RenderLights( );

	//////////////////////////////
	//	render supporting text	//
	//////////////////////////////
	switch ( iOutputLevel ) {
		case 2:
			//Everything
			RenderControlText( );
		case 1:
			//Frame Rate + Device
			RenderDeviceText( );
		case 0:
			//Frame Rate Only
			RenderBasicText( iTrisDrawn );
			break;
		default:
			RenderBasicText( iTrisDrawn );
	}

	if ( (iOutputLevel != 2) && bDisplayStats )
		RenderStatsText();
	
	//tell D3D we've finished sending geometry to be rendered.
	pDev->EndScene();
	
	//tell D3D to present the image we just built on the screen.
	//its at this point that the user will actually see something.
	pDev->Present(NULL,NULL,NULL,NULL);

	return true;

} //OneFrameRender()


/*-------------------------------------------------------------*\

	OneFrameUpdate()
	----------------
	
	This function does all the background work needed to get the
	internal data ready to be rendered. It also does any 
	per-frame calculations needed by the graphics engine.

\*-------------------------------------------------------------*/
bool CGraphics::OneFrameUpdate() {

	//variables for later
	double dFinal;

	//calculate the current camera angle
	if ( bAnimateCamera ) {
		fCamAngle += (((float)( GetTickCount() - dwLastCameraChange )) / 1000.0f) * (float)iCameraSpeed;
		if ( fCamAngle > 360.0f )
			fCamAngle -= 360.0f;
		dwLastCameraChange = GetTickCount();
		D3DXMatrixLookAtLH( &MatView, 
							new D3DXVECTOR3(50 * (float)sin( fCamAngle * (D3DX_PI / 180) ),
											70,
											50 * (float)cos( fCamAngle * (D3DX_PI / 180) )), 
							new D3DXVECTOR3(0,5,0), 
							new D3DXVECTOR3(0,1,0) );
		pDev->SetTransform( D3DTS_VIEW, &MatView );
	}

	//update animated lights
	AnimateLights();

	//PROFILE: fMS_CalcShadowVols
	QueryPerformanceCounter( &liStart );

	//Calculate shadow volumes
#ifdef SHADOW_VOLUME_UPDATE_ALTERNATE

	//NOTE: setting the value here to anything higher than
	//      3 gives noticably poor results. The frame rate
	//		will remain high, but the shadows appear very jerky
	if ( iShadowUpdateSpeed >= 2 ) {
		CalculateShadowVolumes();
		iShadowUpdateSpeed = 0;
	} else
		iShadowUpdateSpeed++;

#else

	CalculateShadowVolumes();

#endif

	//PROFILE: fMS_CalcShadowVols
	QueryPerformanceCounter( &liEnd );
	QueryPerformanceFrequency( &liFreq );
	dFinal = (double)( liEnd.QuadPart - liStart.QuadPart ) / (double)liFreq.QuadPart;
	fMS_CalcShadowVols = (float)( dFinal * 1000 );

	//calculate frame rate, and calculate the more precise time
	QueryPerformanceCounter( &liEnd2 );
	QueryPerformanceFrequency( &liFreq2 );
	dFinal = (double)( liEnd2.QuadPart - liStart2.QuadPart ) / (double)liFreq2.QuadPart;
	fMS_TotalFrame = (float)( dFinal * 1000 );

	QueryPerformanceCounter( &liStart2 );

	if ( ( GetTickCount() - dwLastFPSCheck ) >= iFPSPollRate ) {
		//time to calculate the frame rate
		dwLastFPSCheck = GetTickCount();
		fCurrFPS = (float)iCurrCnt * (1000.0f/(float)iFPSPollRate);
		iCurrCnt = 0;
	}
	iCurrCnt++;

	return true;
}


/*-------------------------------------------------------------*\

	SET: Wireframe(bool)
	GET: Wireframe()
	--------------------

	bWire =	true to enable wireframe, false to disable

	Tells the rendering engine to draw all geometry in a
	wireframe mode.

\*-------------------------------------------------------------*/
void CGraphics::Wireframe(bool bWire) {
	if (!bInitOkay)
		return;

	bWireFrame = bWire;
	if (bWire)
		pDev->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	else
		pDev->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
}

bool CGraphics::Wireframe() {
	return bWireFrame;
}


/*-------------------------------------------------------------*\

	SET: Light(int,bool)
	GET: Light(int)
	--------------------

	idx = index for the light that is to be enabled/disabled
	enabled = true to turn on, false to turn off

	Allows the host application to enable/disable any of the
	lights in the scene.

\*-------------------------------------------------------------*/
void CGraphics::Light( int idx, bool enabled ) {
	if ( !bInitOkay )
		return;

	if ( idx >= 1 && idx <= iLightCount ) {
		lList[idx - 1]->bEnabled = enabled;
		pDev->LightEnable( idx - 1, (BOOL)lList[ idx - 1 ]->bEnabled );
	}

	if ( idx == 0 ) {
		pDev->SetRenderState( D3DRS_AMBIENT,	enabled 
												? AMB_LIGHT
												: D3DCOLOR_XRGB( 0, 0, 0 ) 
							);
		bAmbientLight = enabled;
	}

} //Light( int, bool )


bool CGraphics::Light( int idx ) {
	if ( bInitOkay )
		if ( idx >= 1 && idx <= iLightCount ) {
			return lList[idx - 1]->bEnabled;
		} else if ( idx == 0 ) {
			return bAmbientLight;
		}
		

	return false; //default control path

} //Light( int )


/*-------------------------------------------------------------*\

	SET: AnimateCamera(bool)
	GET: AnimateCamera()
	------------------------

	bAnimate = true to animate, false to pause.

	Allows the host application to turn on/off the rotation
	of the camera.

\*-------------------------------------------------------------*/
void CGraphics::AnimateCamera( bool bAnimate ) {
	bAnimateCamera = bAnimate;
	if ( bAnimateCamera )
		dwLastCameraChange = GetTickCount();
}

bool CGraphics::AnimateCamera() {
	return bAnimateCamera;
}	


/*-------------------------------------------------------------*\

	SET: TextOutput(int)
	GET: TextOutput()
	--------------------

	iLevel = integer value: 0 (basic frame rate)
							1 (device information)
							2 (Control output)

	Allows the host to set how much/what text is displayed
	when each frame is rendered.

\*-------------------------------------------------------------*/
void CGraphics::TextOutput( int iLevel ) {
	iOutputLevel = iLevel;
	if ( iOutputLevel > 2 )
		iOutputLevel = 0;
}

int CGraphics::TextOutput( ) {
	return iOutputLevel;
}


/*-------------------------------------------------------------*\

	SET: ModelDetail(int)
	GET: ModelDetail()
	---------------------

	D = Level of detail, 0 for low, 1 for medium, 2 for high

	Allows, at runtime, for different levels of geometric
	detail to be used.

\*-------------------------------------------------------------*/
void CGraphics::ModelDetail( int D ) {

	if ( D != iModelDetail && ( D >= 0 && D <= 2 ) ) {
		//detail has changed
		iModelDetail = D;
		LoadGeometry( iModelDetail );
	}

}

int CGraphics::ModelDetail( ) {
	return iModelDetail;
}


/*-------------------------------------------------------------*\

	SET: StatisticsOutput(bool)
	GET: StatisticsOutput()
	---------------------------

	bOutput = true to enable output, false to disable

	Outputs some basic profiling statistics on the shadow
	generation/rendering process. Statistics cant be rendered
	if the TextOutput() level is set to 2 (occupies the same
	part of the screen).

\*-------------------------------------------------------------*/
void CGraphics::StatisticsOutput( bool bOutput ) {
	bDisplayStats = bOutput;
}

bool CGraphics::StatisticsOutput( ) {
	return bDisplayStats;
}


/*-------------------------------------------------------------*\

	InitialiseDevice(HWND,int,int,int,bool)
	---------------------------------------

	hWnd	= handle to the target window
	iW		= width of the render target (0 for windowed)
	iH		= height of the render target (0 for windowed)
	iD		= depth of the render target (inherits windows settings in windowed mode)
	bWin	= true for windowed, false for fullscreen

	This is the code that gets the graphics engine rolling, all
	the configuration and creation for Direct3D 9.0 is contained
	in this function.

\*-------------------------------------------------------------*/
bool CGraphics::InitialiseDevice(HWND hWnd, int iW, int iH, int iD, bool bWin) {
	
	pD3D9 = Direct3DCreate9( D3D_SDK_VERSION );
	if (pD3D9 == NULL)
		throw CGraphicsException("Could not create Direct3D9 interface","CGraphics.cpp",__LINE__);

	D3DPRESENT_PARAMETERS pp;
	ZeroMemory( &pp, sizeof(pp) );
	pp.Windowed = bWin;			
	pp.EnableAutoDepthStencil = true;
	pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	pp.BackBufferCount = 1; //double buffer
	if (bWin) {
		D3DDISPLAYMODE *md = new D3DDISPLAYMODE;
		pD3D9->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, md );
		pp.BackBufferFormat = md->Format;
		SAFE_DELETE( md );
	} else {
		pp.BackBufferWidth = iW;
		pp.BackBufferHeight = iH;
		switch (iD) {
			case 16:
				pp.BackBufferFormat = D3DFMT_R5G6B5;
				break;
			case 24:
				pp.BackBufferFormat = D3DFMT_R8G8B8;
				break;
			case 32:
				pp.BackBufferFormat = D3DFMT_X8R8G8B8;
				break;
			default:
				throw CGraphicsException("Incorrect display format specified","CGraphics.cpp",__LINE__);
		}
 	}
	//enumerate depth-stencil formats.
		//D24FS8
		//D24S8
		//D24X4S4
		//D15S1
		//ERROR!
	char *sDepthFmt = new char[12];
	if ( pD3D9->CheckDepthStencilMatch( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, pp.BackBufferFormat, pp.BackBufferFormat, D3DFMT_D24FS8 ) != D3D_OK ) {
		if ( pD3D9->CheckDepthStencilMatch( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, pp.BackBufferFormat, pp.BackBufferFormat, D3DFMT_D24S8 ) != D3D_OK ) {
			if ( pD3D9->CheckDepthStencilMatch( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, pp.BackBufferFormat, pp.BackBufferFormat, D3DFMT_D24X4S4 ) != D3D_OK ) {
				if ( pD3D9->CheckDepthStencilMatch( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, pp.BackBufferFormat, pp.BackBufferFormat, D3DFMT_D24FS8 ) != D3D_OK ) {
					//ERROR!
					throw CGraphicsException( "This demo requires a hardware stencil buffer, but no support exists on this hardware.", __FILE__, __LINE__ );
				} else {
					//support for a d15s1 buffer
					pp.AutoDepthStencilFormat = D3DFMT_D15S1;
					sDepthFmt = "D:15, S:1";
				}
			} else {
				//support for a d24x4s4 buffer
				pp.AutoDepthStencilFormat = D3DFMT_D24X4S4;
				sDepthFmt = "D:24, S:4";
			}
		} else {
			//support for a d24s8 buffer
			pp.AutoDepthStencilFormat = D3DFMT_D24S8;
			sDepthFmt = "D:24, S:8";
		}
	} else {
		//support for a d24fs8 buffer
		pp.AutoDepthStencilFormat = D3DFMT_D24FS8;
		sDepthFmt = "D:24f, S:8";
	}
#ifdef DISPLAY_VSYNC
	pp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;			//vsync is on
#else
	pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;	//vsync is off
#endif
	
	D3DCAPS9 caps;
	DWORD behaviour;
	pD3D9->GetDeviceCaps( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps );
	//we want to select hardware transform and lighting if its available:
	if ( caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT )
		behaviour = D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		behaviour = D3DCREATE_SOFTWARE_VERTEXPROCESSING;

	if( FAILED( pD3D9->CreateDevice( D3DADAPTER_DEFAULT,
									 D3DDEVTYPE_HAL,
									 hWnd,
									 behaviour,
									 &pp,
									 &pDev ) ) )
		throw CGraphicsException("Could not create Direct3D9 device","CGraphics.cpp",__LINE__);

	//retrieve information about the display mode:
	sDispInfo = new char[100];
	char *sFmt = new char[30];
	switch (pp.BackBufferFormat) {
		case D3DFMT_R5G6B5:
			sFmt = "16bit 565";
			iScreenDepth = 16;
			break;
		case D3DFMT_X1R5G5B5:
			sFmt = "16bit X555";
			iScreenDepth = 16;
			break;
		case D3DFMT_R8G8B8:
			sFmt = "24bit 888";
			iScreenDepth = 24;
			break;
		case D3DFMT_X8R8G8B8:
			sFmt = "32bit X888";
			iScreenDepth = 32;
			break;
		case D3DFMT_A8R8G8B8:
			sFmt = "32bit 8888";
			iScreenDepth = 32;
			break;
		default:
			sFmt = "Unknown Format";
			iScreenDepth = 0;
			break;
	}
	sprintf(	sDispInfo, "%dx%d (%s %s), TexMem: %dmb",pp.BackBufferWidth, pp.BackBufferHeight, sFmt, sDepthFmt, pDev->GetAvailableTextureMem() / (1024 * 1024)	);
	iScreenWidth = pp.BackBufferWidth;
	iScreenHeight = pp.BackBufferHeight;
	fmtDepth = pp.AutoDepthStencilFormat;


	D3DADAPTER_IDENTIFIER9 *pID = new D3DADAPTER_IDENTIFIER9;
	pD3D9->GetAdapterIdentifier( D3DADAPTER_DEFAULT, 0, pID);
	sDevInfo = new char[200];
	sprintf(	sDevInfo, "%s (v%d.%d build %d)",pID->Description, 
										   LOWORD(pID->DriverVersion.HighPart),
										   HIWORD(pID->DriverVersion.LowPart),
										   LOWORD(pID->DriverVersion.LowPart)
										   );

	//setup basic render states
	pDev->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
	pDev->SetRenderState( D3DRS_LIGHTING, TRUE );
	pDev->SetRenderState( D3DRS_ZENABLE, TRUE );
	pDev->SetRenderState( D3DRS_AMBIENT, AMB_LIGHT );
	bAmbientLight = true; //internal flag

	//set up texture information
	pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	pDev->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	pDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );

	pDev->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
    pDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    pDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
    pDev->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE );

	//configure stencil modes - saves re-setting them every frame..
	pDev->SetRenderState( D3DRS_STENCILENABLE,		FALSE );
	pDev->SetRenderState( D3DRS_STENCILZFAIL,		D3DSTENCILOP_KEEP );
	pDev->SetRenderState( D3DRS_STENCILFAIL,		D3DSTENCILOP_KEEP );
	pDev->SetRenderState( D3DRS_STENCILREF,			0x1 );
	pDev->SetRenderState( D3DRS_STENCILMASK,		0xffffffff );
	pDev->SetRenderState( D3DRS_STENCILWRITEMASK,	0xffffffff );
	pDev->SetRenderState( D3DRS_CCW_STENCILFUNC,	D3DCMP_ALWAYS );
	pDev->SetRenderState( D3DRS_CCW_STENCILZFAIL,	D3DSTENCILOP_KEEP );
	pDev->SetRenderState( D3DRS_CCW_STENCILFAIL,	D3DSTENCILOP_KEEP );
	pDev->SetRenderState( D3DRS_CCW_STENCILPASS,	D3DSTENCILOP_DECR );
	b2SidedStencils = ( ( caps.StencilCaps & D3DSTENCILCAPS_TWOSIDED ) != 0 );

	//function and _PASS values are the only ones that change each frame
	
	//load relevent textures into memory
	if (!LoadTextures(pp.BackBufferFormat))
		throw CGraphicsException("Could not execute CGraphics::LoadTextures()","CGraphics.cpp",__LINE__);
	
	//load required geometry into memory
	if (!LoadGeometry( iModelDetail ))
		throw CGraphicsException("Could not execute CGraphics::LoadGeometry()","CGraphics.cpp",__LINE__);

	//configure the lights for the scene (4 dynamic)
	if (!Setuplights())
		throw CGraphicsException("Could not execute CGraphics::Setuplights()", __FILE__, __LINE__ );

	fmtDisp = pp.BackBufferFormat;

	
	//allocate memory for shadow volumes
	for ( int i = 0; i < 4; i++ )
		pVols[i] = new ShadowVolume();


	//configure the matrices
	D3DXMatrixIdentity( &MatWorld );
	D3DXMatrixLookAtLH( &MatView, new D3DXVECTOR3(10,50,-70), new D3DXVECTOR3(0,5,0), new D3DXVECTOR3(0,1,0) );
	D3DXMatrixPerspectiveFovLH( &MatProj, D3DX_PI/4, 4/3, 1, 200 );

	pDev->SetTransform ( D3DTS_WORLD, &MatWorld );
	pDev->SetTransform ( D3DTS_VIEW, &MatView );
	pDev->SetTransform ( D3DTS_PROJECTION, &MatProj );

	//configure the fonts
	hMainFont = CreateFont(iFontSize,0,0,0,FW_NORMAL,false,false,0,
						   ANSI_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
						   DEFAULT_QUALITY, DEFAULT_PITCH, "Verdana");
	D3DXCreateFont( pDev, hMainFont, &lpFont );

	//get a base value for the first frames calculations
	dwLastFrameUpdate = GetTickCount();
	
	bInitOkay = true; //we succeeded this far!
	return true;
} //InitialiseDevice()


/*-------------------------------------------------------------*\

	LoadTextures(D3DFORMAT)
	-----------------------

	fmtDisplay = the format of the render target, used to match
				 texture formats with the rendering target..

	This loads the generally required textures for the app. For
	this sample only one texture is required.
	Note: textures for the meshes are loaded seperately.

\*-------------------------------------------------------------*/
bool CGraphics::LoadTextures(D3DFORMAT fmtDisplay) {

	//check for support for A8R8G8B8 textures
	if ( pD3D9->CheckDeviceFormat( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, fmtDisplay, 0, D3DRTYPE_TEXTURE, D3DFMT_A8R8G8B8 ) != D3D_OK )
		throw CGraphicsException( "The device does not support 32bit ARGB textures, cannot continue!", __FILE__, __LINE__ );

	//load the texture
	if ( FAILED( D3DXCreateTextureFromFileEx( pDev, "Graphics/PointLight.dds",
												32, 32, 1, 0, D3DFMT_A8R8G8B8,
												D3DPOOL_MANAGED, 
												D3DX_FILTER_LINEAR, D3DX_FILTER_LINEAR,
												0, NULL, NULL,
												&pLightTex ) ) ) 
	{
		throw CGraphicsException( "Cannot load Graphics/PointLight.dds", __FILE__, __LINE__ );
	}

	return true;
}


/*-------------------------------------------------------------*\

	LoadGeometry()
	--------------

	iLOD = the level of detail (low,med,high) to load

	This code handles the loading of all the geometry from disk.
	The CMesh class handles the majority of the work though.

\*-------------------------------------------------------------*/
bool CGraphics::LoadGeometry( int iLOD ) {

	try {

		//release any existing memory
		for ( int i = 0; i < 8; i++ )
			SAFE_DELETE( mshCaster[ i ] );

		//Create the meshes
		for ( int i = 0; i < 4; i++ )
			mshCaster[ i ] = new CMesh( pDev, "Meshes/Floor.x", "Graphics/" );

		switch ( iLOD ) {
			case 0: //MINIMUM DETAIL LEVEL
				//for ( int i = 4; i < 8; i++ )
				//	mshCaster[ i ] = new CMesh( pDev, "Meshes/Archway_Min.x", "Graphics/" );
				mshCaster[ 4 ] = new CMesh( pDev, "Meshes/Archway_Min.x", "Graphics/" );
				mshCaster[ 5 ] = new CMesh( pDev, "Meshes/Teapot_Lo.x", "Graphics/" );
				mshCaster[ 6 ] = new CMesh( pDev, "Meshes/Links_Lo.x", "Graphics/" );
				mshCaster[ 7 ] = new CMesh( pDev, "Meshes/Knot_Lo.x", "Graphics/" );
				break;

			case 1: //MEDIUM DETAIL LEVEL
				//for ( int i = 4; i < 8; i++ )
				//	mshCaster[ i ] = new CMesh( pDev, "Meshes/Archway_Med.x", "Graphics/" );
				mshCaster[ 4 ] = new CMesh( pDev, "Meshes/Archway_Med.x", "Graphics/" );
				mshCaster[ 5 ] = new CMesh( pDev, "Meshes/Teapot_Md.x", "Graphics/" );
				mshCaster[ 6 ] = new CMesh( pDev, "Meshes/Links_Md.x", "Graphics/" );
				mshCaster[ 7 ] = new CMesh( pDev, "Meshes/Knot_Md.x", "Graphics/" );
				break;

			case 2: //MAXIMUM DETAIL LEVEL
				//for ( int i = 4; i < 8; i++ )
				//	mshCaster[ i ] = new CMesh( pDev, "Meshes/Archway_Max.x", "Graphics/" );
				mshCaster[ 4 ] = new CMesh( pDev, "Meshes/Archway_Max.x", "Graphics/" );
				mshCaster[ 5 ] = new CMesh( pDev, "Meshes/Teapot_Hi.x", "Graphics/" );
				mshCaster[ 6 ] = new CMesh( pDev, "Meshes/Links_Hi.x", "Graphics/" );
				mshCaster[ 7 ] = new CMesh( pDev, "Meshes/Knot_Hi.x", "Graphics/" );
				break;

			default: //DEFAULT TO MINIMUM DETAIL LEVEL
				for ( int i = 4; i < 8; i++ )
					mshCaster[ i ] = new CMesh( pDev, "Meshes/Archway_Min.x", "Graphics/" );

		}

		//Set the parameters
		mshCaster[0]->scale( 1.0f, 1.0f, 1.0f );
		mshCaster[0]->position( -15.0f, 0.0f, -15.0f );

		mshCaster[1]->scale( 1.0f, 1.0f, 1.0f );
		mshCaster[1]->position( -15.0f, 0.0f, 15.0f );

		mshCaster[2]->scale( 1.0f, 1.0f, 1.0f );
		mshCaster[2]->position( 15.0f, 0.0f, -15.0f );

		mshCaster[3]->scale( 1.0f, 1.0f, 1.0f );
		mshCaster[3]->position( 15.0f, 0.0f, 15.0f );

		mshCaster[4]->scale( 0.5f, 0.8f, 0.5f );
		mshCaster[4]->position( -15.0f, 0.0f, 0.0f );
		mshCaster[4]->rotation( 0.0f, 90.0f, 0.0f );

		mshCaster[5]->scale( 0.5f, 0.8f, 0.5f );
		mshCaster[5]->position( 15.0f, 0.0f, 0.0f );
		mshCaster[5]->rotation( 0.0f, 90.0f, 0.0f );

		mshCaster[6]->scale( 0.5f, 0.5f, 0.5f );
		mshCaster[6]->position( 0.0f, 0.0f, -15.0f );

		mshCaster[7]->scale( 0.3f, 0.3f, 0.3f );
		mshCaster[7]->position( 0.0f, 10.0f, 15.0f );

		return true;

	} catch( CGraphicsException e ) {
		e.ReportError();
		return false;

	}
} //LoadGeometry()


/*-------------------------------------------------------------*\

	Setuplights()
	-------------

	This function deals with setting the initial parameters for
	the 4 dynamic lights in this sample program.

\*-------------------------------------------------------------*/
bool CGraphics::Setuplights( ) {

	//GENERAL / AMBIENT
	pDev->SetRenderState( D3DRS_LIGHTING, TRUE );
	pDev->SetRenderState( D3DRS_AMBIENT, AMB_LIGHT );
	
	//RED LIGHT
	SAFE_DELETE( lList[0] );
	lList[0] = new light;
	lList[0]->bAnimated = true;
	lList[0]->bChanged = true;
	lList[0]->bEnabled = true;
	lList[0]->bCastsShadows = true;
	lList[0]->dwLastUpdate = GetTickCount();
	lList[0]->fPos = 0.0f;
	lList[0]->fSpeed = 45.0f;
	lList[0]->vecOldPos = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	ZeroMemory( &lList[0]->lt, sizeof( D3DLIGHT9 ) );
	lList[0]->lt.Type = D3DLIGHT_POINT;
	lList[0]->lt.Position = D3DXVECTOR3( 0.0f, 10.0f, 0.0f );
	lList[0]->lt.Range = 40.0f;
	lList[0]->lt.Attenuation1 = 1.0f / 4.0f;
	lList[0]->lt.Diffuse.a = 1.0f;		lList[0]->lt.Diffuse.r = 2.0f;
	lList[0]->lt.Ambient = lList[0]->lt.Diffuse;
	
	pDev->SetLight( 0, &lList[0]->lt );
	pDev->LightEnable( 0, (BOOL)lList[0]->bEnabled );



	//GREEN LIGHT
	SAFE_DELETE( lList[1] );
	lList[1] = new light;
	lList[1]->bAnimated = true;
	lList[1]->bChanged = true;
	lList[1]->bEnabled = true;
	lList[1]->bCastsShadows = true;
	lList[1]->dwLastUpdate = GetTickCount();
	lList[1]->fPos = 0.0f;
	lList[1]->fSpeed = 90.0f;
	lList[1]->vecOldPos = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	ZeroMemory( &lList[1]->lt, sizeof( D3DLIGHT9 ) );
	lList[1]->lt.Type = D3DLIGHT_POINT;
	lList[1]->lt.Position = D3DXVECTOR3( 0.0f, 10.0f, 0.0f );
	lList[1]->lt.Range = 55.0f;
	lList[1]->lt.Attenuation1 = 1.0f / 1.5f;
	lList[1]->lt.Diffuse.a = 1.0f;		lList[1]->lt.Diffuse.g = 2.0f;
	lList[1]->lt.Ambient = lList[1]->lt.Diffuse;
	
	pDev->SetLight( 1, &lList[1]->lt );
	pDev->LightEnable( 1, (BOOL)lList[1]->bEnabled );

	//BLUE LIGHT
	SAFE_DELETE( lList[2] );
	lList[2] = new light;
	lList[2]->bAnimated = true;
	lList[2]->bChanged = true;
	lList[2]->bEnabled = true;
	lList[2]->bCastsShadows = true;
	lList[2]->dwLastUpdate = GetTickCount();
	lList[2]->fPos = 0.0f;
	lList[2]->fSpeed = 30.0f;
	lList[2]->vecOldPos = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	ZeroMemory( &lList[2]->lt, sizeof( D3DLIGHT9 ) );
	lList[2]->lt.Type = D3DLIGHT_POINT;
	lList[2]->lt.Position = D3DXVECTOR3( 0.0f, 10.0f, 0.0f );
	lList[2]->lt.Range = 20.0f;
	lList[2]->lt.Attenuation1 = 1.0f / 2.0;
	lList[2]->lt.Diffuse.a = 1.0f;		lList[2]->lt.Diffuse.b = 3.0f;
	lList[2]->lt.Ambient = lList[2]->lt.Diffuse;
	
	pDev->SetLight( 2, &lList[2]->lt );
	pDev->LightEnable( 2, (BOOL)lList[2]->bEnabled );

	//WHITE LIGHT
	SAFE_DELETE( lList[3] );
	lList[3] = new light;
	lList[3]->bAnimated = true;
	lList[3]->bChanged = true;
	lList[3]->bEnabled = true;
	lList[3]->bCastsShadows = true;
	lList[3]->dwLastUpdate = GetTickCount();
	lList[3]->fPos = 0.0f;
	lList[3]->fSpeed = 30.0f;
	lList[3]->vecOldPos = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	ZeroMemory( &lList[3]->lt, sizeof( D3DLIGHT9 ) );
	lList[3]->lt.Type = D3DLIGHT_DIRECTIONAL;
	lList[3]->lt.Position = D3DXVECTOR3( 0.0f, 10.0f, 0.0f );
	lList[3]->lt.Direction = D3DXVECTOR3( 1.0f, 1.0f, 1.0f );
	D3DXVec3Normalize( (D3DXVECTOR3*)&lList[3]->lt.Direction, (D3DXVECTOR3*)&lList[3]->lt.Direction );
	lList[3]->lt.Range = 20.0f;
	lList[3]->lt.Attenuation1 = 1.0f / 2.0;
	lList[3]->lt.Diffuse.a = lList[3]->lt.Diffuse.r = lList[3]->lt.Diffuse.g = lList[3]->lt.Diffuse.b = 0.6f;
	lList[3]->lt.Ambient = lList[3]->lt.Diffuse;
	
	pDev->SetLight( 3, &lList[3]->lt );
	pDev->LightEnable( 3, (BOOL)lList[3]->bEnabled );


	return true;

} //Setuplights()


/*-------------------------------------------------------------*\

	AnimateLights()
	---------------

	This code operates on a per-frame basis and uses a variety
	of parametric equations to animate the dynamic lights in
	the scene.

\*-------------------------------------------------------------*/
void CGraphics::AnimateLights( ) {

	//GENERAL
	for ( int i = 0; i < iLightCount; i++ ) {
		lList[i]->fPos += (((float)( GetTickCount() - lList[i]->dwLastUpdate )) / 1000.0f) * lList[i]->fSpeed;
		if ( lList[i]->fPos > 360.0f )
			lList[i]->fPos -= 360.0f;
		lList[i]->dwLastUpdate = GetTickCount();
	}

	//RED LIGHT
	if ( lList[0]->bAnimated ) {
        //Calculate a new position
		lList[0]->lt.Position = D3DXVECTOR3( 20 * (float)cos( lList[0]->fPos * ( D3DX_PI / 180.0f ) ),
											 17.0f + 3.0f * (float)fabs(sin( 3.0f * lList[0]->fPos * ( D3DX_PI / 180.0f ))),
											 20 * (float)sin( lList[0]->fPos * ( D3DX_PI / 180.0f ) )
											 );
		pDev->SetLight( 0, &lList[0]->lt );
		lList[0]->bChanged = true; //we may need to update the shadow volumes on the next pass..
	}

	//GREEN LIGHT
	if ( lList[1]->bAnimated ) {
		//Calculate a new position
		lList[1]->lt.Position = D3DXVECTOR3( 20 * (float)cos( lList[1]->fPos * ( D3DX_PI / 180.0f ) ) * (float)sin( lList[1]->fPos * ( D3DX_PI / 180.0f ) ),
											 16.0f,
											 20 * (float)cos( lList[1]->fPos * ( D3DX_PI / 180.0f ) )
											 );
		pDev->SetLight( 1, &lList[1]->lt );
		lList[1]->bChanged = true; //we may need to update the shadow volumes on the next pass..
	}

	//BLUE LIGHT
	if ( lList[2]->bAnimated ) {
		//Calculate a new position
		lList[2]->lt.Position = D3DXVECTOR3( ( 18.0f + 12.0f * (float)cos( 4.0f * lList[2]->fPos * ( D3DX_PI / 180.0f ) ) ) * (float)sin( lList[2]->fPos * ( D3DX_PI / 180.0f ) ),
											  8.0f,
											 ( 18.0f + 12.0f * (float)cos( 4.0f * lList[2]->fPos * ( D3DX_PI / 180.0f ) ) ) * (float)cos( lList[2]->fPos * ( D3DX_PI / 180.0f ) )
											 );
		pDev->SetLight( 2, &lList[2]->lt );
		lList[2]->bChanged = true; //we may need to update the shadow volumes on the next pass..
	}

	//WHITE LIGHT
	if ( lList[3]->bAnimated ) {
		//Calculate a new position
		lList[3]->lt.Direction = D3DXVECTOR3( (float)cos( lList[3]->fPos * (D3DX_PI / 180.0f ) ),
											  1.0f,
											  (float)(sin( lList[3]->fPos * (D3DX_PI/180.0f) ) )
											 );
		D3DXVec3Normalize( (D3DXVECTOR3*)&lList[3]->lt.Direction, (D3DXVECTOR3*)&lList[3]->lt.Direction );
		pDev->SetLight( 3, &lList[3]->lt );
		lList[3]->bChanged = true; //we may need to update the shadow volumes on the next pass..
	}

} //AnimateLights()


/*-------------------------------------------------------------*\

	CalculateShadowVolumes()
	------------------------

	This function makes sure that all necessary shadow volumes
	have been generated. This code needs to be executed when
	either the lights have moved or the geometry changes.

\*-------------------------------------------------------------*/
void CGraphics::CalculateShadowVolumes() {

	D3DXMATRIXA16 mat;
	D3DXVECTOR3 vecLight;
	int k = 0;
	for ( int i = 0; i < iLightCount; i++ ) {
		if ( lList[i]->bEnabled && lList[i]->bCastsShadows ) {
			//the current light requires that shadow volumes be generated

			for ( int j = 4; j < 8; j++ ) {
				//for each object we transform the light into model space
				if ( lList[ i ]->lt.Type == D3DLIGHT_POINT ) {
					
					//calculate the world->model space transform matrix
					D3DXMatrixInverse( &mat, NULL, &mshCaster[j]->TransformMatrix() );

					//multiply the lights position, to go from world->model space
					vecLight.x = lList[i]->lt.Position.x*mat._11 + lList[i]->lt.Position.y*mat._21 + lList[i]->lt.Position.z*mat._31 + mat._41;
					vecLight.y = lList[i]->lt.Position.x*mat._12 + lList[i]->lt.Position.y*mat._22 + lList[i]->lt.Position.z*mat._32 + mat._42;
					vecLight.z = lList[i]->lt.Position.x*mat._13 + lList[i]->lt.Position.y*mat._23 + lList[i]->lt.Position.z*mat._33 + mat._43;

				} else if ( lList[ i ]->lt.Type == D3DLIGHT_DIRECTIONAL ) {
					//the direction vector needs to be multiplied by the rotation matrix for the current mesh, otherwise
					//the direction will not be properly in model space, instead a hybrid of model and world space..

					//construct the correct rotation matrix
					D3DXMATRIXA16 matRot, matTmp;
					D3DXVECTOR3 pRot;

					pRot = mshCaster[ j ]->rotation( );

					D3DXMatrixIdentity( &matRot );

					D3DXMatrixIdentity( &matTmp );
					D3DXMatrixRotationX( &matTmp, pRot.x * ( D3DX_PI / 180.0f ) );
					D3DXMatrixMultiply( &matRot, &matRot, &matTmp );

					D3DXMatrixIdentity( &matTmp );
					D3DXMatrixRotationY( &matTmp, pRot.y * ( D3DX_PI / 180.0f ) );
					D3DXMatrixMultiply( &matRot, &matRot, &matTmp );

					D3DXMatrixIdentity( &matTmp );
					D3DXMatrixRotationZ( &matTmp, pRot.z * ( D3DX_PI / 180.0f ) );
					D3DXMatrixMultiply( &matRot, &matRot, &matTmp );

					//invert the transform
					D3DXMatrixInverse( &matRot, NULL, &matRot );

					//multiply the direction and the matrix
					vecLight.x = lList[i]->lt.Direction.x*matRot._11 + lList[i]->lt.Direction.y*matRot._21 + lList[i]->lt.Direction.z*matRot._31 + matRot._41;
					vecLight.y = lList[i]->lt.Direction.x*matRot._12 + lList[i]->lt.Direction.y*matRot._22 + lList[i]->lt.Direction.z*matRot._32 + matRot._42;
					vecLight.z = lList[i]->lt.Direction.x*matRot._13 + lList[i]->lt.Direction.y*matRot._23 + lList[i]->lt.Direction.z*matRot._33 + matRot._43;
				}

				//we now request that the volume be generated..
				pVols[ k ]->Reset();
				pVols[ k ]->SetWorldMatrix( mshCaster[ j ]->TransformMatrix() );
				pVols[ k ]->BuildFromMesh( mshCaster[j], vecLight );
				k++;
			}
		}
	}

	iShadowCount = k; //store this value for later use

} //CalculateShadowVolumes()


/*-------------------------------------------------------------*\

	RenderBasicText()
	-----------------

	iTrisDrawn = number of triangles drawn, used to generate
				 some of the statistics

	This uses the standard DrawText() calls to output some
	basic information to the screen.

\*-------------------------------------------------------------*/
void CGraphics::RenderBasicText( int iTrisDrawn ) {

		//required variables/allocations
		RECT rcTextFPS, rcTextHint;
		char *pMsg = new char[100];

		//set the draw area for the frame-rate
		rcTextFPS.right = 400;	rcTextFPS.bottom = 20;
		rcTextFPS.top = 5;		rcTextFPS.left = 5;

		//set the draw area for the text hint
		rcTextHint.top = iScreenHeight - 20;		rcTextHint.left = 5;
		rcTextHint.bottom = iScreenHeight;			rcTextHint.right = 200;

		//all other strings are calculated in InitialiseDevice()
		sprintf( pMsg, "FPS: %d, Tris: %d @ %.2f M/Sec, Shadows: %d", (int)fCurrFPS, iTrisDrawn, ((float)iTrisDrawn * fCurrFPS) / 1000000.0f, iShadowCount );

		//begin drawing the text.
		lpFont->Begin();
			lpFont->DrawTextA(pMsg,-1, &rcTextFPS, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(255,200,0));
			lpFont->DrawTextA( "[T] Toggle Text Information", -1, &rcTextHint, DT_TOP | DT_LEFT, D3DCOLOR_XRGB( 255, 200, 0 ) );
		lpFont->End();

		//clear up.
		delete pMsg;

} //RenderBasicText()


/*-------------------------------------------------------------*\

	RenderDeviceText()
	------------------

	Outputs some information about the created device and
	the driver its using.

\*-------------------------------------------------------------*/
void CGraphics::RenderDeviceText( ) {

	RECT rcTextRes, rcTextDev;

	//set the draw area for the resolution information
	rcTextRes.left = 5;		rcTextRes.top = 22;

	//set the draw area for the device information
	rcTextDev.left = 5;		rcTextDev.top = 39;		

	//begin drawing the text.
	lpFont->Begin();
		lpFont->DrawTextA(sDispInfo,-1, &rcTextRes, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(255,200,0));
		lpFont->DrawTextA(sDevInfo,-1, &rcTextDev, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(255,200,0));
	lpFont->End();


} //RenderDeviceText()


/*-------------------------------------------------------------*\

	RenderControlText()
	-------------------

	Outputs a list of the controls and their current states
	to the screen. Allows the user to know what buttons do what!

\*-------------------------------------------------------------*/
void CGraphics::RenderControlText( ) {

	RECT rcText;
	char* pMsg = NULL;
	lpFont->Begin();
		rcText.left = 5;

		rcText.top = 3 + (8 * (iFontSize + 4 ));
		pMsg = bAmbientLight 
				? "[1] - Disable Grey Ambient Light"
				: "[1] - Enable Grey Ambient Light";
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(128,128,128) );

		rcText.top = 3 + (9 * (iFontSize + 4 ));
		pMsg = lList[ 0 ]->bEnabled 
							? "[2] - Disable Red Point Light"
							: "[2] - Enable Red Point Light";
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(255,128,128) );

		rcText.top = 3 + (10 * (iFontSize + 4 ));
		pMsg = lList[ 1 ]->bEnabled 
							? "[3] - Disable Green Point Light"
							: "[3] - Enable Green Point Light";
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(128,255,128) );

		rcText.top = 3 + (11 * (iFontSize + 4 ));
		pMsg = lList[ 2 ]->bEnabled 
							? "[4] - Disable Blue Point Light"
							: "[4] - Enable Blue Point Light";
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(128,128,255) );

		rcText.top = 3 + (12 * (iFontSize + 4 ));
		pMsg = lList[ 3 ]->bEnabled 
							? "[5] - Disable White Directional Light"
							: "[5] - Enable White Directional Light";
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(255,255,255) );

		rcText.top = 3 + (13 * (iFontSize + 4 ));
		pMsg = new char[200];
		sprintf( pMsg, "[LEFT] or [RIGHT] - Alter Model Detail (LOD: %d, TRIS: %d)\0", iModelDetail, mshCaster[ 4 ]->trianglesInMesh() );
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );
		delete pMsg;

		rcText.top = 3 + (14 * (iFontSize + 4 ));
		lpFont->DrawTextA("[S] - Toggle Statistics Output", -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );

		rcText.top = 3 + (15 * (iFontSize + 4 ));
		lpFont->DrawTextA("[P] - Toggle Camera Movement", -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );

		rcText.top = 3 + (16 * (iFontSize + 4 ));
		lpFont->DrawTextA("[W] - Toggle Wireframe Rendering", -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );
		
	lpFont->End();

} //RenderControlText()


/*-------------------------------------------------------------*\

	RenderStatsText()
	-----------------

	Outputs the simple profiling results for the shadow volume
	creation and rendering process.

\*-------------------------------------------------------------*/
void CGraphics::RenderStatsText( ) {

	RECT rcText;
	char* pMsg = NULL;
	lpFont->Begin();

		rcText.left = 5;
		
		//TOTAL FRAME TIME
		pMsg = new char[200];
		sprintf( pMsg, "Total Frame Time: %.3fms (100%%)", fMS_TotalFrame );
		rcText.top = 3 + (4 * (iFontSize + 4 ));
		lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );
		delete pMsg;

			//CREATE SHADOWS
			pMsg = new char[200];
			sprintf( pMsg, "   CalculateShadowVolumes(): %.3fms (%.3f%%)", fMS_CalcShadowVols, ( 100.0f * fMS_CalcShadowVols ) / fMS_TotalFrame );
			rcText.top = 3 + (5 * (iFontSize + 4 ));
			lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );
			delete pMsg;

			//RENDER SHADOW LIGHTS
			pMsg = new char[200];
			sprintf( pMsg, "   Render Shadows/lights: %.3fms (%.3f%%)", fMS_RenderShadows, ( 100.0f * fMS_RenderShadows ) / fMS_TotalFrame );
			rcText.top = 3 + (6 * (iFontSize + 4 ));
			lpFont->DrawTextA( pMsg, -1, &rcText, DT_TOP|DT_LEFT, D3DCOLOR_XRGB(200,128,64) );
			delete pMsg;

	lpFont->End();

} //RenderStatsText()


/*-------------------------------------------------------------*\

	RenderLights()
	--------------

	Renders the lights to the screen in the form of point
	sprites.

\*-------------------------------------------------------------*/
void CGraphics::RenderLights( ) {
	
	//0. Any Variables
	LPDIRECT3DVERTEXBUFFER9 pBuf = NULL;
	UINT iLights = 0;
	UINT idx = 0;
	D3DXMATRIXA16 matIdent;

	//2. Create vertex buffer
	for ( int i = 0; i < 3; i++ )
		iLights += lList[i]->bEnabled ? 1 : 0;

	if ( iLights == 0 )
		return; //no lights to render!

	if ( FAILED( pDev->CreateVertexBuffer( iLights * sizeof( vertexParticle ), D3DUSAGE_POINTS | D3DUSAGE_WRITEONLY, FVF_VERTEXPARTICLE, D3DPOOL_MANAGED, &pBuf, NULL ) ) )
		throw CGraphicsException( "(CGraphics::RenderLights) Could not create vertex buffer!", __FILE__ , __LINE__ );
    
	//3. Fill vertex buffer
	vertexParticle *pVerts = NULL;
	pBuf->Lock( 0, 0, (void**)&pVerts, 0 );
		
	for ( int i = 0; i < 3; i++ ) {
		if ( lList[i]->bEnabled ) {
			//add a vertex of the correct color and position
			pVerts[idx].p = lList[i]->lt.Position;
			pVerts[idx].c =	D3DCOLOR_ARGB(	(int)(lList[i]->lt.Diffuse.a * 255),
											(int)(lList[i]->lt.Diffuse.r * 255),
											(int)(lList[i]->lt.Diffuse.g * 255),
											(int)(lList[i]->lt.Diffuse.b * 255)
											);
			idx++;
		}
	}

	pBuf->Unlock();

	//4. enable psprite rendering
	pDev->SetRenderState( D3DRS_POINTSCALEENABLE, FALSE );
	pDev->SetRenderState( D3DRS_POINTSPRITEENABLE, TRUE );
	pDev->SetRenderState( D3DRS_POINTSIZE, FtoDW( 15.0f ) );
	pDev->SetRenderState( D3DRS_POINTSIZE_MIN, FtoDW(0.00f) );
    pDev->SetRenderState( D3DRS_POINTSCALE_A,  FtoDW(0.00f) );
    pDev->SetRenderState( D3DRS_POINTSCALE_B,  FtoDW(0.00f) );
    pDev->SetRenderState( D3DRS_POINTSCALE_C,  FtoDW(1.00f) );
	pDev->SetRenderState( D3DRS_LIGHTING, FALSE );
	pDev->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	pDev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
	D3DXMatrixIdentity( &matIdent );

	//5. render
	pDev->SetFVF( FVF_VERTEXPARTICLE );
	pDev->SetTexture( 0, pLightTex );
	pDev->SetTransform( D3DTS_WORLD, &matIdent );
	pDev->SetStreamSource( 0, pBuf, 0, sizeof( vertexParticle ) );
	pDev->DrawPrimitive( D3DPT_POINTLIST, 0, iLights );

	//6. reset renderstates
	pDev->SetRenderState( D3DRS_LIGHTING, TRUE );
	pDev->SetRenderState( D3DRS_POINTSPRITEENABLE, FALSE );
	pDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

	//7. release vertex buffer
	SAFE_RELEASE( pBuf );

} //RenderLights()