/*
	
	TITLE:		CGraphics.h
	DESC:		Interface definition header for simple Direct3D9 sample program
	DATE:		January 2003

	AUTHOR:		Jack Hoxley

*/


#ifndef INC_CGRAPHICS_H	//check for inclusion guard
#define INC_CGRAPHICS_H	//declare inclusion guard

//include necessary system headers:
#include <math.h>
#include <d3dx9.h>
#include "CMesh.h"
#include "ShadowCalculator.h"

inline DWORD FtoDW( float f ) { return *((DWORD*)&f); }

class CGraphics {	
private:
	//--------------//
	//	VARIABLES	//
	//--------------//

	//core objects
	LPDIRECT3D9 pD3D9;
	LPDIRECT3DDEVICE9 pDev;

	//matrices
	D3DXMATRIXA16 MatWorld;
	D3DXMATRIXA16 MatView;
	D3DXMATRIXA16 MatProj;

	//geometry
	CMesh *mshCaster[8]; //4*floor + 4*archways

	//vertex structs
	struct vertex3D {
		D3DXVECTOR3 p;	//position
		D3DXVECTOR2 t;	//texture
	};
	static const DWORD FVF_VERTEX3D =  D3DFVF_XYZ|D3DFVF_TEX1;

	struct vertex2D {
		D3DXVECTOR4 p;	//position
		DWORD c;		//color
		D3DXVECTOR2 t;	//texture
	};
	static const DWORD FVF_VERTEX2D = D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1;

	struct vertexParticle {
		D3DXVECTOR3 p;
		DWORD c;
	};
	static const DWORD FVF_VERTEXPARTICLE = D3DFVF_XYZ|D3DFVF_DIFFUSE;

	//lights
	struct light {
		D3DLIGHT9 lt;
		D3DXVECTOR3 vecOldPos; //the position of the light on the previous frame (for delta calculation)
		bool bEnabled;
		bool bAnimated;
		bool bChanged;
		bool bCastsShadows;
		int iMode; //0=off, 1=on (static), 2=on (animated)
		float fPos;
		float fSpeed;
		DWORD dwLastUpdate;
	};

	static const int iLightCount = 4;
	light *lList[ iLightCount ];		// r,g and b point lights
	ShadowVolume *pVols[ iLightCount * 4 ]; //4 casting objects
	int iShadowCount; //how many shadow volumes exist for this frame..
	LPDIRECT3DTEXTURE9 pLightTex;
	bool bAmbientLight;

	//fonts
	HFONT hMainFont;
	LPD3DXFONT lpFont;

	//per-frame variables
	DWORD dwLastFrameUpdate;

	//constants
	static const int iFPSPollRate = 200; //200ms / 5x a second
	static const int iFontSize = 14;
	static const int iCameraSpeed = 45; //45 degrees/second

	//control variables
	int iOutputLevel;	//amount of on-screen text
	bool bWireFrame;
	bool bUseTextures;
	bool bCulling;
	bool bDrawFPS;
	bool bAnimateCamera;

	//misc variables
	bool bInitOkay;
	DWORD dwLastFPSCheck;
	float fCurrFPS;
    int iCurrCnt;
	char *sDevInfo;
	char *sDispInfo;
	float fCamAngle;
	DWORD dwLastCameraChange;
	bool b2SidedStencils;
	int iModelDetail;
	int iShadowUpdateSpeed;

	//statistics
	LARGE_INTEGER liStart, liEnd, liFreq;
	LARGE_INTEGER liStart2, liEnd2, liFreq2; //for the fMS_TotalFrame timer
	float fMS_TotalFrame;
	float fMS_RenderShadows;
	float fMS_CalcShadowVols;
	bool bDisplayStats;

	//reference variables
	int iScreenWidth;
	int iScreenHeight;
	int iScreenDepth;
	D3DFORMAT fmtDisp;
	D3DFORMAT fmtDepth;

	//--------------//
	//	FUNCTIONS	//
	//--------------//
	void ResetAllVars();
	bool InitialiseDevice(HWND hWnd, int iW, int iH, int iD, bool bWin);
	bool LoadTextures(D3DFORMAT fmtDisplay);
	bool LoadGeometry( int iLOD );
	bool Setuplights();
	void AnimateLights();
	void CalculateShadowVolumes();
	void RenderBasicText( int iTrisDrawn );		//frame rate
	void RenderDeviceText();					//device/display mode
	void RenderControlText();					//what buttons do what
	void RenderStatsText();						//outputs some timing values
	void RenderLights();

public:
	//----------------------------------//
	//	CONSTRUCTORS AND DESTRUCTORS	//
	//----------------------------------//
	CGraphics(HWND hWnd);							//windowed mode constructor
	CGraphics(HWND hWnd, int iW, int iH, int iD);	//fullscreen mode constructor
	~CGraphics();									//destructor

	//----------------------//
	//	PUBLIC INTERFACES	//
	//----------------------//
	static bool IsDisplayModeOkay(int iW, int iH, int iD);	//checks the parameters and determines if the
													//device supports it as a fullscreen resolution.
	bool OneFrameUpdate();							//updates maths/physics for current frame
	bool OneFrameRender();							//renders the current frame

	//--------------//
	//	PROPERTIES	//
	//--------------//
	void Wireframe(bool bWire);					//set
	bool Wireframe();							//get

	void Light( int idx, bool enabled );		//set
	bool Light( int idx );						//get

	void AnimateCamera( bool bAnimate );		//set
	bool AnimateCamera();						//get

	void TextOutput( int iLevel );				//set
	int TextOutput( );							//get

	void StatisticsOutput( bool bOutput );		//set
	bool StatisticsOutput( );					//get

	void ModelDetail( int D );					//set
	int ModelDetail( );							//get

};

#endif	//#ifndef INC_CGRAPHICS_H