/*
 *	CMesh.cpp
 *	Jack Hoxley
 *
 */

#include "CMesh.h"
#include <d3dx9.h>
#include <math.h>
#include <stdio.h>
#include "CGraphicsException.h"
#include "ShadowCalculator.h"


//--------------------------------------------------------
//	CMesh()
//	default constructor, deals with default settings for
//	all internal variables and structures
//--------------------------------------------------------
CMesh::CMesh( ) {

	pMsh = NULL;
	pMat = NULL;
	pTex = NULL;
	dwNumMaterials = 0L;

	vecPosition = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	vecScale = D3DXVECTOR3( 1.0f, 1.0f, 1.0f );
	vecRotate = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );

	D3DXMatrixIdentity( &matMesh );
	bRebuildMatrices = true;

} //ctor()


//--------------------------------------------------------
//	CMesh( dev, char, char)
//	extended class constructor, does basic initialisation
//	and then passes execution to Load() to get the data
//	into memory.
//--------------------------------------------------------
CMesh::CMesh( LPDIRECT3DDEVICE9 dev, char *sFilename, char *sTexPath ) {

	pMsh = NULL;
	pMat = NULL;
	pTex = NULL;
	dwNumMaterials = 0L;

	vecPosition = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	vecScale = D3DXVECTOR3( 1.0f, 1.0f, 1.0f );
	vecRotate = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );

	D3DXMatrixIdentity( &matMesh );
	bRebuildMatrices = true;

	Load( dev, sFilename, sTexPath ); //load the mesh into memory

} //ctor(dev,char,char)


//--------------------------------------------------------
//	~CMesh()
//	default destructor for the class, makes sure all used
//	data is cleaned up properly.
//--------------------------------------------------------
CMesh::~CMesh( ) {

	SAFE_RELEASE( pMsh );
	SAFE_DELETE_ARRAY( pMat );

	for ( DWORD i = 0; i < dwNumMaterials; i++ )
		SAFE_RELEASE( pTex[i] );

} //dtor


//--------------------------------------------------------
//	Load()
//	loads the specified file into memory, including
//	any necessary textures and materials.
//--------------------------------------------------------
void CMesh::Load( LPDIRECT3DDEVICE9 dev, char *sFilename, char *sTexPath ) {

	//variables
	LPD3DXBUFFER pD3DXBuf;
	D3DXMATERIAL *matList;

	//load the data into memory
	if ( FAILED( D3DXLoadMeshFromX( sFilename, D3DXMESH_SYSTEMMEM, dev, NULL, &pD3DXBuf, NULL, &dwNumMaterials, &pMsh ) ) )
		throw CGraphicsException( "Unable to load mesh from file (D3DXLoadMeshFromX failed)", __FILE__, __LINE__ );

	//configure the materials and textures
	matList = ( D3DXMATERIAL* )pD3DXBuf->GetBufferPointer();
	pMat = new D3DMATERIAL9[ dwNumMaterials ];
	pTex = new LPDIRECT3DTEXTURE9[ dwNumMaterials ];
	char *sFile = new char[256];

	for ( DWORD i = 0; i < dwNumMaterials; i++ ) {
		pMat[i] = matList[i].MatD3D;
		pMat[i].Ambient = pMat[i].Diffuse;

		pTex[i] = NULL;
		if ( matList[i].pTextureFilename != NULL && lstrlen( matList[i].pTextureFilename ) > 0 ) {
			sprintf( sFile, "%s%s", sTexPath, matList[i].pTextureFilename );

			if ( FAILED( D3DXCreateTextureFromFile( dev, sFile, &pTex[i] ) ) ) {
				throw CGraphicsException( "Could not load texture for mesh (D3DXCreateTextureFromFile failed)", __FILE__, __LINE__ );
			}
			
		}
	}

	SAFE_DELETE( sFile );

	//calculate the bounding sphere of the mesh...
	LPD3DXMESH p;
	pMsh->CloneMeshFVF( 0, D3DFVF_XYZ, dev, &p );
	
	LPDIRECT3DVERTEXBUFFER9 pVB;
	p->GetVertexBuffer( &pVB );
	
	D3DXVECTOR3 *vecP;
	pVB->Lock( 0, 0, (void**)&vecP, 0 );

		//now we calculate the bounding sphere
		float xTotal = 0.0f, yTotal = 0.0f, zTotal = 0.0f;

		for ( DWORD i = 0; i < p->GetNumVertices(); i++ ) {
			xTotal += vecP[ i ].x;
			yTotal += vecP[ i ].y;
			zTotal += vecP[ i ].z;
		}
		vecCentre = D3DXVECTOR3( xTotal / p->GetNumVertices(),
								 yTotal / p->GetNumVertices(),
								 zTotal / p->GetNumVertices() );

		fSphereSize = 0.0f;
		for ( DWORD i = 0; i < p->GetNumVertices(); i++ ) {
			float f = (float)sqrt(  ( vecP[i].x - vecCentre.x ) * ( vecP[i].x - vecCentre.x ) +
									( vecP[i].y - vecCentre.y ) * ( vecP[i].y - vecCentre.y ) +
									( vecP[i].z - vecCentre.z ) * ( vecP[i].z - vecCentre.z ) 
								 );

			if ( f > fSphereSize )
				fSphereSize = f;
		}

	pVB->Unlock( );
	
	SAFE_RELEASE( p );

	//reset the internal matrices
	vecPosition = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
	vecScale = D3DXVECTOR3( 1.0f, 1.0f, 1.0f );
	vecRotate = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );

	SAFE_RELEASE( pD3DXBuf );

} //Load()



//--------------------------------------------------------
//	Render()
//	render the stored mesh onto the specified device
//	if the matrix params have been altered since the
//	last pass then they're recalculated
//--------------------------------------------------------
void CMesh::Render( LPDIRECT3DDEVICE9 dev ) {

	if ( bRebuildMatrices ) {
		// rebuild the matrix
		//use an optimized method to generate the matrix
		//2.6x faster than using the chained D3DX commands
		float CosRx = (float)cos( vecRotate.x * ( D3DX_PI / 180 ) );
		float CosRy = (float)cos( vecRotate.y * ( D3DX_PI / 180 ) );
		float CosRz = (float)cos( vecRotate.z * ( D3DX_PI / 180 ) );

		float SinRx = (float)sin( vecRotate.x * ( D3DX_PI / 180 ) );
		float SinRy = (float)sin( vecRotate.y * ( D3DX_PI / 180 ) );
		float SinRz = (float)sin( vecRotate.z * ( D3DX_PI / 180 ) );

		matMesh._11 = ( vecScale.x * CosRy * CosRz );
		matMesh._12 = ( vecScale.x * CosRy * SinRz );
		matMesh._13 = -( vecScale.x * SinRy );
		matMesh._14 = 0.0f;

		matMesh._21 = -(vecScale.y * CosRx * SinRz) + (vecScale.y * SinRx * SinRy * CosRz);
		matMesh._22 = (vecScale.y * CosRx * CosRz) + (vecScale.y * SinRx * SinRy * SinRz);
		matMesh._23 = (vecScale.y * SinRx * CosRy);
		matMesh._24 = 0.0f;

		matMesh._31 = (vecScale.z * SinRx * SinRz) + (vecScale.z * CosRx * SinRy * CosRz);
		matMesh._32 = -(vecScale.z * SinRx * CosRx) + (vecScale.z * CosRx * SinRy * SinRz);
		matMesh._33 = (vecScale.z * CosRx * CosRy);
		matMesh._34 = 0.0f;

		matMesh._41 = vecPosition.x;
		matMesh._42 = vecPosition.y;
		matMesh._43 = vecPosition.z;
		matMesh._44 = 1.0f;

		//set a flag to make sure its only rebuilt when necessary
		bRebuildMatrices = false;
	}

	//Render the mesh
	dev->SetTransform( D3DTS_WORLD, &matMesh );
	for ( DWORD i = 0; i < dwNumMaterials; i++ ) {
		dev->SetMaterial( &pMat[i] );
		dev->SetTexture( 0, pTex[i] );
		pMsh->DrawSubset( i );
	}

} //Render()


//--------------------------------------------------------
//	trianglesInMesh()
//	returns the number of triangles, useful for
//	calculating rendering statistics
//--------------------------------------------------------
int CMesh::trianglesInMesh( ) {

	if ( pMsh )
		return (int)pMsh->GetNumFaces();
	else
		return 0;

} //trianglesInMesh()


//--------------------------------------------------------
//	verticesInMesh()
//	returns the number of vertices, useful for
//	calculating rendering statistics
//--------------------------------------------------------
int CMesh::verticesInMesh( ) {

	if ( pMsh )
		return (int)pMsh->GetNumVertices();
	else
		return 0;

} //verticesInMesh()


//--------------------------------------------------------
//	getMesh()
//	Some algorithms require direct access to the mesh
//	in these cases they can use this function to get
//	the raw internal data
//--------------------------------------------------------
LPD3DXMESH CMesh::getMesh( ) {
	
	if ( pMsh )
		return pMsh;
	else
		return NULL;

} //getMesh()


//--------------------------------------------------------
//	getBoundingSphereCentre()
//	This returns the centre of the current geometry
//	combined with the next function, gives the mesh's
//	bounding sphere
//--------------------------------------------------------
D3DXVECTOR3 CMesh::getBoundingSphereCentre( ) {
	
	return vecCentre;

} //getBoundingSphereCentre()


//--------------------------------------------------------
//	getBoundingSphereSize()
//	This will return the maximum size of the geometry,
//	scaled according to the scale vector. Very useful 
//	for preliminary influence/interaction testing.
//--------------------------------------------------------
float CMesh::getBoundingSphereSize( ) {

	return fSphereSize;
	
} //getBoundingSphereSize()


//--------------------------------------------------------
//	TransformMatrix(Matrix)
//	Allows the owning app. to specify a different matrix
//	to that generated internally
//--------------------------------------------------------
void CMesh::TransformMatrix( D3DXMATRIX mat ) {
	
	matMesh = (D3DXMATRIXA16)mat;
	bRebuildMatrices = false; //don't overwrite it

} //TransformMatrix(Matrix)


//--------------------------------------------------------
//	TransformMatrix()
//	returns the current matrix, or generates the latest
//	one if its out of date
//--------------------------------------------------------
D3DXMATRIX CMesh::TransformMatrix( ) {
	
	if ( bRebuildMatrices ) {
		float CosRx = (float)cos( vecRotate.x * ( D3DX_PI / 180 ) );
		float CosRy = (float)cos( vecRotate.y * ( D3DX_PI / 180 ) );
		float CosRz = (float)cos( vecRotate.z * ( D3DX_PI / 180 ) );

		float SinRx = (float)sin( vecRotate.x * ( D3DX_PI / 180 ) );
		float SinRy = (float)sin( vecRotate.y * ( D3DX_PI / 180 ) );
		float SinRz = (float)sin( vecRotate.z * ( D3DX_PI / 180 ) );

		matMesh._11 = ( vecScale.x * CosRy * CosRz );
		matMesh._12 = ( vecScale.x * CosRy * SinRz );
		matMesh._13 = -( vecScale.x * SinRy );
		matMesh._14 = 0.0f;

		matMesh._21 = -(vecScale.y * CosRx * SinRz) + (vecScale.y * SinRx * SinRy * CosRz);
		matMesh._22 = (vecScale.y * CosRx * CosRz) + (vecScale.y * SinRx * SinRy * SinRz);
		matMesh._23 = (vecScale.y * SinRx * CosRy);
		matMesh._24 = 0.0f;

		matMesh._31 = (vecScale.z * SinRx * SinRz) + (vecScale.z * CosRx * SinRy * CosRz);
		matMesh._32 = -(vecScale.z * SinRx * CosRx) + (vecScale.z * CosRx * SinRy * SinRz);
		matMesh._33 = (vecScale.z * CosRx * CosRy);
		matMesh._34 = 0.0f;

		matMesh._41 = vecPosition.x;
		matMesh._42 = vecPosition.y;
		matMesh._43 = vecPosition.z;
		matMesh._44 = 1.0f;

		//set a flag to make sure its only rebuilt when necessary
		bRebuildMatrices = false;
		return (D3DXMATRIX)matMesh;

	} else {
		return (D3DXMATRIX)matMesh;

	}

} //TransformMatrix()


//--------------------------------------------------------
//	position
//	Various methods for allowing the owner to set
//	the mesh's position in world space
//--------------------------------------------------------
void CMesh::positionX( float x ) {
	if ( x != vecPosition.x )
		bRebuildMatrices = true;
	vecPosition.x = x;
}

float CMesh::positionX( ) {
	return vecPosition.x;
}

void CMesh::positionY( float y ) {
	if ( y != vecPosition.y )
		bRebuildMatrices = true;
	vecPosition.y = y;
}

float CMesh::positionY( ) {
	return vecPosition.y;
}

void CMesh::positionZ( float z ) {
	if ( z != vecPosition.z )
		bRebuildMatrices = true;
	vecPosition.z = z;
}

float CMesh::positionZ( ) {
	return vecPosition.z;
}

void CMesh::position( float x, float y, float z ) {
	vecPosition.x = x;
	vecPosition.y = y;
	vecPosition.z = z;
}

D3DXVECTOR3 CMesh::position( ) {
	return vecPosition;
}



//--------------------------------------------------------
//	scaling
//	Various methods for allowing the owner to set
//	the mesh's scale in world space
//--------------------------------------------------------
void CMesh::scaleX( float x ) {
	if ( x != vecScale.x )
		bRebuildMatrices = true;
	vecScale.x = x;
}

float CMesh::scaleX( ) {
	return vecScale.x;
}

void CMesh::scaleY( float y ) {
	if ( y != vecScale.y )
		bRebuildMatrices = true;
	vecScale.y = y;
}

float CMesh::scaleY( ) {
	return vecScale.y;
}

void CMesh::scaleZ( float z ) {
	if ( z != vecScale.z )
		bRebuildMatrices = true;
	vecScale.z = z;
}

float CMesh::scaleZ( ) {
	return vecScale.z;
}

void CMesh::scale( float x, float y, float z ) {
	vecScale.x = x;
	vecScale.y = y;
	vecScale.z = z;
}

D3DXVECTOR3 CMesh::scale( ) {
	return vecScale;
}


//--------------------------------------------------------
//	rotation
//	Various methods for allowing the owner to set
//	the mesh's rotation (in degrees) in world space
//--------------------------------------------------------
void CMesh::rotationX( float x ) {
	if ( x != vecRotate.x )
		bRebuildMatrices = true;
	vecRotate.x = x;
}

float CMesh::rotationX( ) {
	return vecRotate.x;
}

void CMesh::rotationY( float y ) {
	if ( y != vecRotate.y )
		bRebuildMatrices = true;
	vecRotate.y = y;
}

float CMesh::rotationY( ) {
	return vecRotate.y;
}

void CMesh::rotationZ( float z ) {
	if ( z != vecRotate.z )
		bRebuildMatrices = true;
	vecRotate.z = z;
}

float CMesh::rotationZ( ) {
	return vecRotate.z;
}

void CMesh::rotation( float x, float y, float z ) {
	vecRotate.x = x;
	vecRotate.y = y;
	vecRotate.z = z;
}

D3DXVECTOR3 CMesh::rotation( ) {
	return vecRotate;
}